/*
 * GAI - General Applet Interface Library
 * Copyright (C) 2003 Jonas Aaberg <cja@gmx.net>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * 
 *             Dedicated to Evelyn Reimann - Min ss sv gp af!!
 */

#include <stdio.h>
#include <time.h>
#include "config.h"
#include "gai.h"
#include "gai-private.h"


#define GAI (*gai_instance)

#define GAI_debug (gai_instance && GAI.debug)

static int depth = 0;
static const char spaces[] = "                                              ";

void 
gai_log_msg (const char * function, const char *msg)
{
    FILE *df;
    if (! GAI_debug) return;
    df = fopen("/tmp/gai-debug-output","a");
    if (df <= 0) return;
    fprintf(df, "%.*s%s: %s\n",depth,spaces,function,msg);
    fflush(df);
    fclose(df);
}

void 
gai_log_int (const char* function, int i)
{
    FILE *df;
    if(! GAI_debug) return;
    df = fopen("/tmp/gai-debug-output","a");
    if (df <= 0) return;
    fprintf (df,"%.*s%s: %d\n",depth,spaces,function,i);
    fflush (df);
    fclose (df);
}

void 
gai_log_enter (const char *function)
{
    FILE *df;
    if (! GAI_debug) return;
    df = fopen("/tmp/gai-debug-output","a");
    if (df <= 0) return;
    fprintf (df,"%.*s%s: ... enter\n",depth++,spaces,function);
    fflush (df);
    fclose (df);

}

void 
gai_log_leave (const char *function)
{
    FILE *df;
    if (! GAI_debug) return;
    df = fopen("/tmp/gai-debug-output","a");
    if (df <= 0) return;
    fprintf (df,"%.*s%s: --- leave\n",--depth,spaces,function);
    fflush (df);
    fclose (df);
    if (depth < 0) depth = 0;
}

void 
gai_log_check (const char *function)
{
    FILE *df;
    if (! GAI_debug) return;
    df = fopen("/tmp/gai-debug-output","a");
    if (df <= 0) return;
    fprintf (df,"%.*s%s: !!! checkpoint\n",--depth,spaces,function);
    fflush (df);
    fclose (df);
    if (depth < 0) depth = 0;
}

void 
gai_log_debug_init(void)
{
    struct tm *my_tm;
    time_t my_t;
    FILE *df;
    if(! GAI_debug) return;

    

    my_t = time (NULL);
    my_tm = localtime (&my_t);
    
    df = fopen("/tmp/gai-debug-output","w");
    if(df <= 0) return;
    fprintf (df,"\n*** %s starting - %s ***\n",
	     asctime (my_tm), GAI.applet.name);
    fflush (df);
    fclose (df);
}


void 
gai_display_error_quit(const char *str)
{
    GtkWidget *w;
    GAI_NOTE (str);
    g_assert(str !=NULL);

    w = gtk_message_dialog_new (NULL,0,GTK_MESSAGE_ERROR,GTK_BUTTONS_OK,str);
    gtk_widget_show_all (w);
    g_signal_connect_swapped (G_OBJECT (w), "response",
			      G_CALLBACK (gtk_exit), G_OBJECT (w));

    if (! gai_instance || ! GAI.init_done)
	gtk_main();
}

void 
gai_display_error_continue(const char *str)
{
    GtkWidget *w;
    g_assert(str !=NULL);

    GAI_NOTE (str);
    w = gtk_message_dialog_new (NULL,0,GTK_MESSAGE_ERROR,GTK_BUTTONS_OK,str);
    gtk_widget_show_all (w);
    g_signal_connect_swapped (G_OBJECT (w), "response",
			      G_CALLBACK (gtk_widget_destroy), G_OBJECT (w));

    //    if (! gai_instance || ! GAI.init_done)
    //	gtk_main();
}
