/*
 * GAI - General Applet Interface Library
 * Copyright (C) 2003 Jonas Aaberg <cja@gmx.net>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * 
 *             Dedicated to Evelyn Reimann. - Min ss sv gp af!!
 */

#include "gai.h"
#include "gai-private.h"

#include <gdk/gdkx.h>
#include <gtk/gtk.h>
#include <stdio.h>
#include <unistd.h>

#ifdef GL
#include <GL/gl.h>
#include <GL/glu.h>
#endif



#define GAI (*gai_instance)


static void 
gai_dockapp_window(void)
{
    XSizeHints sizehints;
    Display *xdisplay;
    XWMHints wmhints;
    Window xwin;
    GdkGeometry geo;

    GAI_ENTER;

    GAI.iconwin = NULL;
    GAI.about = NULL;

    GAI.widget = gtk_window_new (GTK_WINDOW_TOPLEVEL);

    gtk_window_set_has_frame (GTK_WINDOW(GAI.widget), FALSE);

    gtk_window_set_wmclass (GTK_WINDOW(GAI.widget),
			    GAI.applet.name, GAI.applet.name);

    gtk_widget_set_events(GAI.widget,GAI.mask);

    gtk_window_set_default_size(GTK_WINDOW(GAI.widget),
				GAI.default_width, GAI.default_height);

    GAI.width = GAI.default_width;
    GAI.height = GAI.default_height;

    GAI.drawingarea = gtk_drawing_area_new();

#ifdef GL
    if(GAI.open_gl){
	gtk_widget_set_gl_capability(GAI.drawingarea,
				     GAI.glconfig,
				     NULL, TRUE,
				     GDK_GL_RGBA_TYPE);
    }
#endif


    /* Fix this strange icon window stuff for window maker */
    if (GAI.window_maker)
    {
	gtk_widget_realize(GAI.widget);

	/* Override the old widget */
	GAI.old_widget = GAI.widget;
	GAI.widget = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_widget_realize(GAI.widget);

	GAI.iconwin = GAI.widget->window;

	gdk_window_reparent(GAI.widget->window, GAI.old_widget->window,0,0);
	gdk_window_set_title(GAI.widget->window, GAI.applet.name);
	geo.min_width = geo.max_width = GAI.default_width;
	geo.min_height = geo.max_height = GAI.default_height;

	gdk_window_set_geometry_hints(GAI.widget->window,& geo,
				      GDK_HINT_MIN_SIZE | GDK_HINT_MAX_SIZE);

	gdk_window_set_events (GAI.widget->window,GAI.mask);

    }

    gtk_container_add(GTK_CONTAINER(GAI.widget), GAI.drawingarea);

    gtk_widget_realize(GAI.drawingarea);

    GAI.window = GAI.drawingarea->window;

    if(!GAI.window_maker)
		gtk_widget_realize(GAI.widget);

    gdk_window_set_type_hint(GAI.widget->window,GDK_WINDOW_TYPE_HINT_DOCK);

    gai_load_background();

    if (! GAI.broken_wm)
    {
	if(GAI.window_maker){
	    xwin = GDK_WINDOW_XWINDOW(GAI.old_widget->window);
	    xdisplay = GDK_WINDOW_XDISPLAY(GAI.old_widget->window);
	    wmhints.icon_window = GDK_WINDOW_XWINDOW(GAI.iconwin);

	}
	else {
	    xdisplay = GDK_WINDOW_XDISPLAY(GAI.widget->window);
	    xwin = GDK_WINDOW_XWINDOW(GAI.widget->window);
	    wmhints.icon_window = xwin;
	}

	sizehints.flags = USSize;
	sizehints.width = GAI.default_width;
	sizehints.height = GAI.default_height;

	XSetWMNormalHints (xdisplay, xwin, &sizehints);

	wmhints.initial_state = WithdrawnState;
	wmhints.icon_x = 0;
	wmhints.icon_y = 0;
	wmhints.window_group = xwin;
	wmhints.flags = 
	    StateHint | IconWindowHint | IconPositionHint | WindowGroupHint;
	XSetWMHints (GDK_WINDOW_XDISPLAY(GAI.widget->window), xwin, &wmhints);

    }
    GAI_LEAVE;


}


int 
gai_timer(gpointer d)
{
#ifdef GL
    GdkGLContext *glcontext;
    GdkGLDrawable *gldrawable=NULL;
#endif

    GAI_ENTER;
    GAI.timer_started = 1;
    if (! GAI.restarting)
    {
#ifdef GL

	if(GAI.open_gl){
	    glcontext = gtk_widget_get_gl_context (GAI.drawingarea);
	    gldrawable = gtk_widget_get_gl_drawable (GAI.drawingarea);

	    if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
		return TRUE;
	}
#endif
	if (GAI.on_update_callback)
	    GAI.on_update_callback(GAI.on_update_userdata);

#ifdef GL
	if(GAI.open_gl){
	    if(gdk_gl_drawable_is_double_buffered(gldrawable))
		gdk_gl_drawable_swap_buffers(gldrawable);
	    else
		glFlush();

	    gdk_gl_drawable_gl_end(gldrawable);
	}
#endif

    }

    GAI_LEAVE;
    return TRUE;
}

static int 
on_mouse_click_callback(GtkWidget *widget, GdkEventButton *e, gpointer d)
{
    int x,y;
    GAI_ENTER;

    if (e->button==3)
    {
	if (GAI.applet_type == GAI_DOCKAPP)
	{
	    gtk_menu_popup (GTK_MENU(GAI.menu), NULL, NULL, NULL, NULL, 
			    e->button, e->time);
	    return TRUE;
	}
	GAI_LEAVE;
	return FALSE;
    }


    if (GAI.rotate && GAI.orient == GAI_VERTICAL)
    {
	x=gai_scale((int)e->y);
	y=gai_scale((int)e->x);
    }else{
	x=gai_scale((int)e->x);
	y=gai_scale((int)e->y);
    }

    if (e->button==1)
    {  
	if (GAI.on_mouse_click1_callback)
	    GAI.on_mouse_click1_callback(x,y, 
					 GAI.on_mouse_click1_userdata);
    }

    if (e->button==2)
    {  
	if (GAI.on_mouse_click2_callback)
	    GAI.on_mouse_click2_callback(x,y, 
					 GAI.on_mouse_click2_userdata);
    }
    
    GAI_LEAVE;    
    return TRUE;
}

static gboolean 
on_scroll_buttons_callback (GtkWidget *widget, GdkEventScroll *event, 
			    gpointer d)
{
    GAI_ENTER;
    if (GAI.on_scroll_buttons_callback)
	GAI.on_scroll_buttons_callback(event->direction,
				       GAI.on_scroll_buttons_userdata);
    GAI_LEAVE;
    return TRUE;
}

static gboolean 
on_enter_callback (GtkWidget *widget, gpointer d)
{
    GAI_ENTER;
    if (GAI.on_enter_callback)
	GAI.on_enter_callback(GAI.on_enter_userdata);
    GAI_LEAVE;
    return TRUE;
}

static gboolean 
on_leave_callback (GtkWidget *widget, gpointer d)
{
    GAI_ENTER;
    if (GAI.on_leave_callback)
	GAI.on_leave_callback (GAI.on_leave_userdata);
    GAI_LEAVE;
    return TRUE;
}

static gboolean 
on_keypress_callback (GtkWidget *widget, GdkEventKey *ev, gpointer d)
{
    GAI_ENTER;
    if (GAI.on_keypress_callback)
	GAI.on_keypress_callback(ev->keyval, GAI.on_keypress_userdata);
    GAI_LEAVE;
    return TRUE;
}

static gboolean 
on_mouse_release_callback (GtkWidget *widget, GdkEventButton *e, gpointer d)
{
    int x,y;
    GAI_ENTER;

    if (e->button==3)
	return FALSE;


    if (GAI.rotate && GAI.orient == GAI_VERTICAL)
    {
	x=gai_scale((int)e->y);
	y=gai_scale((int)e->x);
    }else{
	x=gai_scale((int)e->x);
	y=gai_scale((int)e->y);
    }


    if (e->button==1)
    {  
	if (GAI.on_mouse_release1_callback)
	    GAI.on_mouse_release1_callback(x,y, 
					 GAI.on_mouse_release1_userdata);
    }

    if (e->button==2)
    {  
	if (GAI.on_mouse_release2_callback)
	    GAI.on_mouse_release2_callback(x,y, 
					 GAI.on_mouse_release2_userdata);
    }

    GAI_LEAVE;
    return TRUE;
}


int gai_dies (gpointer d)
{
    GAI_ENTER;

    if (GAI.timer)
	gtk_timeout_remove(GAI.timer);
    
    gtk_main_quit();
    GAI_LEAVE;

    return TRUE;
}

void gai_hook(void)
{

    GAI_ENTER;

    /* gai_data = gai_get_struct_ptr(); */

    g_signal_connect(G_OBJECT(GAI.widget), "button_press_event", 
		      G_CALLBACK(on_mouse_click_callback), NULL);

    g_signal_connect(G_OBJECT (GAI.widget), "delete_event",
		     G_CALLBACK(gai_dies), NULL);

    g_signal_connect(G_OBJECT(GAI.widget), "destroy",
		     G_CALLBACK(gai_dies), NULL);

    if (GAI.on_mouse_release1_callback || 
	GAI.on_mouse_release2_callback)
	g_signal_connect (G_OBJECT(GAI.widget), "button_release_event", 
			  G_CALLBACK(on_mouse_release_callback), NULL);

    if (GAI.on_scroll_buttons_callback)
	g_signal_connect (G_OBJECT(GAI.widget), "scroll_event",
			  G_CALLBACK(on_scroll_buttons_callback), NULL);

    if(GAI.on_keypress_callback)
	g_signal_connect (G_OBJECT(GAI.widget), "key_press_event",
			  G_CALLBACK(on_keypress_callback), NULL);


    if (GAI.on_enter_callback)
	g_signal_connect (G_OBJECT(GAI.widget), "enter_notify_event",
			  G_CALLBACK(on_enter_callback), NULL);

    /* Shall "move_focus_out_of_applet" be used instead of this ?? */
    if (GAI.on_leave_callback)
	g_signal_connect (G_OBJECT(GAI.widget), "leave_notify_event",
			  G_CALLBACK(on_leave_callback), NULL);


    if (GAI.on_update_callback)
	GAI.timer = gtk_timeout_add (GAI.update_interval, gai_timer, NULL);
    else
	GAI.timer = -1;


    if (GAI.on_update_callback)
	GAI.on_update_callback(GAI.on_update_userdata);
    GAI_LEAVE;

}

static void 
gai_simple_about_ok(GtkWidget *widget, gpointer d)
{
    GAI_ENTER;
    gtk_widget_destroy (GAI.about);
    GAI.about = NULL;
    GAI_LEAVE;

} 

static void gai_simple_about(void)
{

    GtkWidget *vbox;
    GtkWidget *label, *separator, *about_ok_button;

    char *tmp;
    GdkPixbuf *pixbuf;

    GAI_ENTER;

    if (GAI.about !=NULL){
	gtk_window_present(GTK_WINDOW(GAI.about));
	GAI_LEAVE;
	return;
    }

    pixbuf = gdk_pixbuf_new_from_file(GAI.applet.icon, NULL);


    GAI.about = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_container_set_border_width(GTK_CONTAINER(GAI.about), 5);
    gtk_window_set_resizable(GTK_WINDOW(GAI.about), FALSE);

    tmp = g_strdup_printf ("About - %s", GAI.applet.nice_name);
    gtk_window_set_title (GTK_WINDOW(GAI.about), tmp);
    g_free (tmp);


    vbox = gtk_vbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER(GAI.about), vbox);

    if(pixbuf!=NULL){
	gtk_box_pack_start (GTK_BOX(vbox), gtk_image_new_from_pixbuf(pixbuf) , TRUE, TRUE, 5);
	g_object_unref(pixbuf);
    }

    


    label = gtk_label_new (NULL);
    tmp = g_strdup_printf("<span size=\"x-large\" font_desc=\"Arial\"><b>" 
		        "%s %s</b></span>",GAI.applet.nice_name, GAI.applet.version);
    gtk_label_set_markup(GTK_LABEL(label), tmp);
    g_free(tmp);
    gtk_label_set_justify(GTK_LABEL(label),GTK_JUSTIFY_CENTER);
    gtk_box_pack_start (GTK_BOX(vbox), label, FALSE, FALSE, 5);
    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
    gtk_misc_set_padding (GTK_MISC(label), 5, 0);

    label = gtk_label_new (GAI.applet.description);
    gtk_label_set_justify(GTK_LABEL(label),GTK_JUSTIFY_CENTER);
    gtk_box_pack_start (GTK_BOX(vbox), label, FALSE, FALSE, 5);
    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
    gtk_misc_set_padding (GTK_MISC(label), 5, 0);

    tmp = g_strdup_printf ("Written by %s",GAI.applet.author);
    label = gtk_label_new (tmp);
    gtk_label_set_justify(GTK_LABEL(label),GTK_JUSTIFY_CENTER);
    g_free (tmp);
    gtk_box_pack_start (GTK_BOX(vbox), label, FALSE, FALSE, 5);
    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
    gtk_misc_set_padding (GTK_MISC(label), 5, 0);


    label = gtk_label_new (NULL);
    tmp = g_strdup_printf("<tt>%s</tt>",GAI.applet.license);
    gtk_label_set_markup(GTK_LABEL(label),tmp);
    gtk_label_set_justify(GTK_LABEL(label),GTK_JUSTIFY_CENTER);
    g_free(tmp);
    gtk_box_pack_start (GTK_BOX(vbox), label, FALSE, FALSE, 5);
    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
    gtk_misc_set_padding (GTK_MISC(label), 5, 0);

    separator = gtk_hseparator_new();
    gtk_widget_show (separator);
    gtk_box_pack_start (GTK_BOX(vbox), separator, TRUE, TRUE, 0);

    about_ok_button = gtk_button_new_from_stock ("gtk-ok");

    gtk_box_pack_start (GTK_BOX(vbox), about_ok_button, FALSE, FALSE, 5);
    gtk_container_set_border_width (GTK_CONTAINER(about_ok_button), 5);
    
    g_signal_connect ((gpointer)about_ok_button, "clicked",
		      G_CALLBACK(gai_simple_about_ok), NULL);

    gtk_widget_show_all (GAI.about);
    GAI_LEAVE;

}


void 
gai_on_preferences_activate (GtkMenuItem *menuitem, gpointer d)
{
    GAI_ENTER;

    if (GAI.have_preferences == 1)
    {
	if (GAI.on_preferences_callback)
	    GAI.on_preferences_callback(GAI.on_preferences_userdata);
    }
    else if (GAI.have_preferences == 2)
    {
	gai_make_preference_window(GAI.applet.name, GAI.gn);
    }
    GAI_ENTER;

}


void 
gai_on_help_activate(GtkMenuItem *menuitem, gpointer d)
{
    GAI_ENTER;
    gai_menu_show_help_text (NULL, NULL);
    GAI_LEAVE;
}


void 
gai_on_about_activate(GtkMenuItem *menuitem, gpointer d)
{
    GAI_ENTER;
#ifdef GAI_WITH_GNOME
    if (GAI.applet_type == GAI_GNOME1 ||
	GAI.applet_type == GAI_GNOME2)
	gai_gnome_about_show();
    else
#endif
	gai_simple_about();
    GAI_LEAVE;
}

void 
gai_on_remove_activate (GtkMenuItem *menuitem, gpointer d)
{
    GAI_ENTER;
    GAI.did_exit_function = 1;
    if (GAI.timer)
	gtk_timeout_remove(GAI.timer);

    if (GAI.on_exit_callback)
	GAI.on_exit_callback (GAI.on_exit_userdata);

    gtk_main_quit();
    GAI_LEAVE;
}

void 
gai_menu_show_help_text(GtkWidget *widget, gpointer d)
{
    GtkWidget *w;

    GAI_ENTER;

    if(GAI.menu_help_text==NULL)
	return;

    w = gtk_message_dialog_new (
	NULL, 0, GTK_MESSAGE_INFO, GTK_BUTTONS_OK, GAI.menu_help_text);
    g_signal_connect_swapped (G_OBJECT(w), "response",
			      G_CALLBACK(gtk_widget_destroy),
			      G_OBJECT(w));
    gtk_widget_show (w);
    GAI_LEAVE;
}


static void 
gai_create_menu(void)
{
    GtkWidget *menu_line=NULL;
    GtkAccelGroup *accel_group;
    int i;

    GAI_ENTER;

    accel_group = gtk_accel_group_new ();
    GAI.menu = gtk_menu_new ();

    for(i=0;i<GAI.menu_entries;i++){

	if(g_array_index(GAI.menu_type,int,i) == GAI_MENU_STOCK){
	    menu_line = gtk_image_menu_item_new_with_mnemonic(g_ptr_array_index(GAI.menu_name,i));

	    if(g_ptr_array_index(GAI.menu_icon,i) != NULL)
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM(menu_line), 
					       gtk_image_new_from_stock (g_ptr_array_index(GAI.menu_icon,i),
									 GTK_ICON_SIZE_MENU));
	    if(g_ptr_array_index(GAI.menu_func,i) !=NULL)
		g_signal_connect (G_OBJECT(menu_line), "activate",
				  G_CALLBACK (g_ptr_array_index(GAI.menu_func,i)), 
				  g_ptr_array_index(GAI.menu_ptr,i));

	}
	if(g_array_index(GAI.menu_type,int,i) == GAI_MENU_SEPARATOR){
	    menu_line = gtk_menu_item_new();
	    gtk_widget_set_sensitive (menu_line, FALSE);
	}

	gtk_container_add (GTK_CONTAINER(GAI.menu), menu_line);
    }
    
    gtk_menu_set_accel_group(GTK_MENU(GAI.menu), accel_group);
    
    gtk_widget_show_all(GAI.menu);

    /* Destroy the arrays, they are of no use now */
    g_array_free(GAI.menu_type, TRUE);
    g_ptr_array_free(GAI.menu_name,TRUE);
    g_ptr_array_free(GAI.menu_icon,TRUE);
    g_ptr_array_free(GAI.menu_func,TRUE);
    g_ptr_array_free(GAI.menu_ptr,TRUE);

    GAI_LEAVE;
}


void gai_dockapp_main(void)
{
#ifdef GL
    GdkGLContext *glcontext;
    GdkGLDrawable *gldrawable;
#endif
    /* gai_data = gai_get_struct_ptr(); */


    GAI_ENTER;

    GAI.scale = 1.0;
    GAI.width = GAI.default_width;
    GAI.height = GAI.default_height;

    gai_dockapp_window();
    gai_hook();
    gai_create_menu();


#ifdef GL
    if(GAI.open_gl){

	glcontext = gtk_widget_get_gl_context (GAI.drawingarea);
	gldrawable = gtk_widget_get_gl_drawable (GAI.drawingarea);

	while(!gdk_gl_drawable_gl_begin (gldrawable, glcontext)){
	    usleep(10000);
	}

	if(GAI.gl_init_func)
	    GAI.gl_init_func(NULL);

	gdk_gl_drawable_gl_end (gldrawable);

    }
#endif 


    gtk_widget_show_all(GAI.widget);

    if(GAI.window_maker)
	  gtk_widget_show_all(GAI.old_widget);


    if (GAI.hide_mouse_ptr)
	gai_hide_mouse_ptr ();
    else
	gai_show_mouse_ptr ();

    /* Start with updating */
    if (GAI.on_update_callback)
	GAI.on_update_callback(GAI.on_update_userdata);

    gtk_main();

    GAI_LEAVE;

}

